from __future__ import annotations

import asyncio
from collections.abc import Hashable
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any, Callable, TypeVar
from weakref import ref

import discord
from discord.ext import commands

from .db import AbstractConnection, AbstractDbFactory

T = TypeVar("T")


class Reserved:
    def __init__(self, value: Any, empty: asyncio.Future) -> None:
        self.value = value
        self.__empty = empty

    async def empty(self) -> None:
        await self.__empty


class ReservedContainer:
    def __init__(self, value: Any, empty: asyncio.Future) -> None:
        self.reserved = Reserved(value, empty)
        self.empty = empty
        self.rc = 0


class Reservation:
    def __init__(
        self,
        container: ref[dict[Hashable, Reservation]],
        key: Hashable,
        loop: asyncio.AbstractEventLoop,
        factory: Callable[[], Any],
    ) -> None:
        self.__container = container
        self.__key = key
        self.__rc: ReservedContainer | None = None
        self.__create_future = loop.create_future
        self.__factory = factory

    def __enter__(self) -> Reserved:
        container = self.__container()
        if container is None:
            raise RuntimeError("⭐ can't reserve in a non-existent container ⭐")
        if self.__rc is None:
            other = container.get(self.__key)
            if other is not None:
                assert other.__rc
                self.__rc = other.__rc
            else:
                self.__rc = ReservedContainer(self.__factory(), self.__create_future())
                container[self.__key] = self
        rc = self.__rc
        rc.rc += 1
        return rc.reserved

    def __exit__(self, et, ev, tb, /):
        rc = self.__rc
        assert rc
        assert rc.rc > 0
        rc.rc -= 1
        if rc.rc == 0:
            rc.empty.set_result(None)
            container = self.__container()
            if container is not None:
                del container[self.__key]


class RefableDict(dict):
    pass


class Reservations:
    def __init__(self) -> None:
        self.__container: dict[Hashable, Reservation] = RefableDict()
        self.__loop = asyncio.get_running_loop()

    def reserve(self, key: Hashable, factory: Callable[[], Any]) -> Reservation:
        return Reservation(ref(self.__container), key, self.__loop, factory)


class StarState:
    def __init__(self, connection: AbstractConnection) -> None:
        self.connection = connection
        self.reservations = Reservations()


@asynccontextmanager
async def state_manager(dbf: AbstractDbFactory):
    async with dbf.from_path(Path("/app/data/stars.db")) as connection:
        yield StarState(connection)


class StarBot(commands.Bot):
    def __init__(self, state: StarState) -> None:
        super().__init__(
            command_prefix="⭐",
            intents=discord.Intents(message_content=True, guild_messages=True, guild_reactions=True, guilds=True),
        )
        self.starstate: StarState = state
